require 'strscan'

##
# This module contains the classes which implement parsers and, when included,
# provides class methods for defining custom parsers.

module MetaParse

  ##
  # Extend including class with ClassMethods.
  
  def self.included(base)
    base.extend ClassMethods
  end

  ##
  # Parse string using an instance method previously defined by MetaParse::ClassMethodss::match_method.
  
  def parse_with_method(method_name, string)
    self.send(method_name, string.meta(self), self)
  end

  module ClassMethods
    ##
    # Defines a method which takes a scanner from provided block.
    # The block passed should return a Matcher or Matcher spec, which is compiled to a Matcher if necessary.
    # The result of calling the defined method is the same as calling match? on the resulting Matcher.

    def match_method(name, &block)
      match_spec = yield
      matcher = MetaParse::Matcher.compile(match_spec)
      
      define_matcher_method(name, matcher)
    end

    ##
    # Returns a RepetitionMatcher which matches the pattern specified by its arguments zero-or-more-times.

    def rep(*args, &block)
      Matcher.compile([:*, *args])
    end

    ##
    # Return a SequentialMatcher.  If block is supplied, it defines a function which will be passed an array
    # of all matched values and which should return a non-nil result for the match as a whole.
    # Sequential matching with terminal block is the mechanism by which arbitrary values can be produced at
    # any step in parsing.
    
    def seq(*args, &block)
      if block_given?
        wrapped = lambda { |scanner, context|
          result = block.call context.matches
          context.matches = []
          result
        }
        args << wrapped
      end
      Matcher.compile([:and, *args])
    end

    ##
    # Return an AlternativeMatcher.
    
    def alt(*args)
      Matcher.compile([:or, *args])
    end

    ##
    # Compile a Matcher spec into corresponding matcher.
    
    def comp(spec)
      Matcher.compile(spec)
    end

    private

    ##
    # Define a match method using supplied Matcher.
    # Used by ::match_method, which should be used instead.

    def define_matcher_method(name, matcher)
      self.send(:define_method, name) do |scanner_spec, context=nil|
        scanner = scanner_spec.meta
        scanner.parser ||= self
        matcher.match scanner, context
      end
    end
  end

  class MetaScanner < StringScanner
    attr_accessor :parser

    ##
    # Initialize MetaScanner with supplied string as input.
    # Optionally-supplied parser will be receiver of Symbol message sent by subclass, FunctionMatcher.
    def initialize(string, parser=nil)
      super string
      @parser = parser
    end

    ##
    # MetaScanner is already a MetaScanner
    def meta
      self
    end

    ##
    # Match and return a single character or return nil, updating internal position on match.
    
    # NOTE: This is a special case and could actually be handled by match_string if necessary.
    def match_char(char)
      c = peek(1)
      if c == char
        self.pos += 1
        c
      end
    end

    ##
    #  Match and return a string or return nil, updating internal position on match.
    
    def match_string(str, position2=0)
      if (string.equal_at(pos, str, position2))
        self.pos += str.length
        str
      end
    end

    ##
    # Scan for a Regexp or string, returning any match, or nil, and updating internal position on match.
    
    def scan(spec)
      case spec
      when Regexp
        result = super spec
        matched
      when String
        if spec.length > 0
          match_string(spec)
        else
          match_char(spec)
        end
      end
    end
  end

  ##
  # General class to match a simple pattern against a scanner.
  # Subclasses implement compound matching.
  
  class Matcher
    attr_accessor :spec

    ##
    # Compile a Matcher specification into a concrete Matcher or subclass.
    
    def self.compile(spec)
      case spec
      when Matcher
        spec
      when String, Regexp
        Matcher.new spec
      when Array
        case spec[0]
        when :or
            compiled_body = spec[1..-1].map { |x| self.compile x }
          AlternativeMatcher.new(compiled_body)
        when :and
            compiled_body = spec[1..-1].map { |x| self.compile x }
          SequentialMatcher.new(compiled_body)
        when :*, :+, :'?'
          compiled_body = self.compile(spec[1])
          case spec[0]
          when :'?'
            min = 0
            max = 1
          when :+
            min = 1
          end
          RepetitionMatcher.new(compiled_body, min, max, *spec[4..-1])
        end
      when Proc, Symbol
        FunctionMatcher.new spec
      end
    end

    ##
    # Initialize Matcher with Matcher spec.
    
    def initialize(spec)
      @spec = spec
    end

    def show
        spec.inspect
    end

    def inspect
      "<match #{show}>"
    end

    ##
    # Try to match own pattern/combination against supplied scanner. Context is unused.
    #
    # Subclasses should implement match? but not call it on peers.
    
    def match?(scanner, context=nil)
      case scanner
      when MetaScanner
        result = scanner.scan spec
        result
      else
        raise "match? requires scanner"
        # FIXME: Why doesn't the coercion below work?
      # when String
      #   match?(MetaParse::MetaScanner.new(scanner), context)
      #   (MetaScanner.new(scanner)).scan spec
      end
    end

    ##
    # Like match? but clone self first if stateful.
    #
    # Subclasses should not implement match, but should call it on peers.
    
    def match(scanner, context=nil)
      (stateful ? clone : self).match? scanner, context
    end

    ##
    # Is this Matcher stateful? A class attribute which may be overriden by subclasses.
    
    def stateful
      false
    end

    ##
    # Syntactic sugar to create MetaScanner and match? self against string.
    
    def m?(string)
      match? MetaScanner.new(string)
    end

    ##
    # Syntactic sugar to create MetaScanner and match self against string.

    def m(string)
      match MetaScanner.new(string)
    end    
  end

  ##
  # Matcher subclass matching sub_match repeatedly.
  
  class RepetitionMatcher < Matcher
    # Match at least min times, if min is present.
    attr_accessor :min

    # Match at most max times, if max is present.   
    attr_accessor :max

    # A proc used to combine match results as by inject, if present.
    attr_accessor :reducer
    
    def initialize(sub_match, min=0, max=nil, reducer=nil, initial_value=[])
      @spec, @min, @max, @reducer, @initial_value = sub_match, min, max, reducer, initial_value
    end

    ##
    # RepetitionMatcher is stateful.
    
    def stateful
      true
    end


    # def match?(scanner, context=nil)
    #   matches = []
    #   while (!max || (matches.count < max)) && (match = spec.match(scanner))
    #     matches << match
    #   end
      
    #   unless min && (matches.count < min)
    #     case finalizer
    #     when Proc
    #       finalizer.call(matches, *finalizer_args)
    #     when Symbol
    #       send(finalizer, matches, *finalizer_args)
    #     when nil 
    #       matches
    #     end
    #   end
    # end
    
    def match?(scanner, context=nil)
      # Need to copy the initial value since it is potentially destructively modified (if an array, for example).
      acc = begin @initial_value.dup
            rescue TypeError
              @initial_value
            end
      match_count = 0

      while (!max || (match_count < max)) && (one_match = spec.match(scanner))
        match_count += 1
        if reducer
          acc = reducer.call(acc, one_match)
        else
          acc << one_match
        end
      end
      unless min && (match_count < min)
        acc
      end
    end

    def show
      "[#{min}, #{max}] #{ spec.show }"
    end
  end

  ##
  # Matcher subclass matching first of alternative submatches.
  
  class AlternativeMatcher < Matcher
    def match?(scanner, context=nil)
      spec.each do |alternative|
        result = alternative.match(scanner)
        return result if result
      end
      return nil
    end
 
    def show
      "first of: (#{ (spec.map &:show).join ', ' })"
    end
  end

  ##
  # Matcher subclass matching submatches sequentially.
  
  class SequentialMatcher < Matcher
    attr_accessor :matches


    ##
    # SequentialMatcher is stateful.
    
    def stateful
      true
    end

    ##
    # Match submatch patterns against scanner sequentially, accumulating results in @matches.
    
    def match?(scanner, context = nil)
      @matches = []
      initial_position = scanner.pos
      last_match = nil

      spec.each do |element|
        last_match = element.match(scanner, self)
        unless last_match
          scanner.pos = initial_position
          return nil
        end
        @matches << last_match
      end
      return last_match
    end

    ##
    # Remove most recent sequential match from result stack (@matches), returning it (pop).
    
    def pop
      @matches.pop
    end

    ##
    # Add (push) value to result stack (@matches).
    
    def push(value)
      @matches.push(value)
    end

    ##
    # Reset result stack (@matches) to be empty.
    
    def clear
      @matches = []
    end

    def show
      "sequentially: (#{ (spec.map &:show).join ', ' })"
    end
  end  

  ##
  # Arbitrary predicate, particularly useful for generating final return value from SequentialMatchers.
  # Context is the containing matcher, and in the case of SequentialMatcher includes access to accumulated matches.
  class FunctionMatcher < Matcher
    include MetaParse

    ##
    # Try to match against scanner using arbitrary predicated specified as Proc or Symbol naming method of Matcher's parser.
    
    def match?(scanner, context=nil)
      case spec
      when Proc
        spec.call(scanner, context)
      when Symbol
        scanner.parser.send spec, scanner, context
      end
    end
  end

  ##
  # Extract matches from context and return them, resetting context's matches to be empty.
  
  def all_matches(scanner, context)
    matches_so_far = context.matches
    context.matches = []
    matches_so_far
  end

  ##
  # Join all matches as extracted by #all_matches.
  
  def all_matches_joined(scanner, context)
    all_matches(scanner,context).join
  end

  ##
  # Joins's string args. Helper function for use in assembling parse results.
  
  def join_strings(array, *args)
    array.join(*args)
  end

  ##
  # Apply block to context's matches stack, then clear context, returning result.
  
  def collapse(&block)
    lambda { |scanner, context|
      stack = context.matches
      result = block.call(stack)
      context.clear
      result
    }
  end
end

class String

  ##
  # Returns true iff receiver (String) is equal to string at index, position, of receiver.
  # If position2 is supplied, begin comparison of string at that index position.
  
  def equal_at(position, string, position2=0)
    for i in position2..(position2 + string.length - 1) do
      return nil unless self[position] == string[i] 
      position += 1
    end
    return true
  end

  ##
  # Return scanner using receiver (String) as source.

  def meta(parser=nil)
    MetaParse::MetaScanner.new(self, parser)
  end
end
